@@ -1,9 +1,19 @@
 
 
-#if !HAVE_TIMEGM
+#include <stdio.h>
+
+#include <yara/mem.h>
+#include <yara/integers.h>
+
+#if defined(WIN32)
+#include <string.h>
+#define strncasecmp _strnicmp
+#define timegm _mkgmtime
+#endif
+
+#if !HAVE_TIMEGM && !defined(WIN32)
 
 #include <time.h>
-#include <stdint.h>
 
 static int is_leap(
     unsigned int year)
