@@ -1,17 +1,30 @@
 /*
 Copyright (c) 2014. The YARA Authors. All Rights Reserved.
 
-Licensed under the Apache License, Version 2.0 (the "License");
-you may not use this file except in compliance with the License.
-You may obtain a copy of the License at
+Redistribution and use in source and binary forms, with or without modification,
+are permitted provided that the following conditions are met:
 
-   http://www.apache.org/licenses/LICENSE-2.0
+1. Redistributions of source code must retain the above copyright notice, this
+list of conditions and the following disclaimer.
 
-Unless required by applicable law or agreed to in writing, software
-distributed under the License is distributed on an "AS IS" BASIS,
-WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-See the License for the specific language governing permissions and
-limitations under the License.
+2. Redistributions in binary form must reproduce the above copyright notice,
+this list of conditions and the following disclaimer in the documentation and/or
+other materials provided with the distribution.
+
+3. Neither the name of the copyright holder nor the names of its contributors
+may be used to endorse or promote products derived from this software without
+specific prior written permission.
+
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
+ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
+ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 
@@ -22,8 +35,8 @@ limitations under the License.
 #include <string.h>
 #include <math.h>
 
-#if _WIN32
-#define PRIu64 "%I64d"
+#if _WIN32 || __CYGWIN__
+#define PRIu64 "I64d"
 #else
 #include <inttypes.h>
 #endif
@@ -42,7 +55,10 @@ int yr_object_create(
     YR_OBJECT** object)
 {
   YR_OBJECT* obj;
-  size_t object_size;
+  int i;
+  size_t object_size = 0;
+
+  assert(parent != NULL || object != NULL);
 
   switch (type)
   {
@@ -111,7 +127,7 @@ int yr_object_create(
       break;
     case OBJECT_TYPE_FUNCTION:
       ((YR_OBJECT_FUNCTION*) obj)->return_obj = NULL;
-      for (int i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
+      for (i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
       {
         ((YR_OBJECT_FUNCTION*) obj)->prototypes[i].arguments_fmt = NULL;
         ((YR_OBJECT_FUNCTION*) obj)->prototypes[i].code = NULL;
@@ -150,6 +166,10 @@ int yr_object_create(
       case OBJECT_TYPE_DICTIONARY:
         ((YR_OBJECT_DICTIONARY*) parent)->prototype_item = obj;
         break;
+
+      case OBJECT_TYPE_FUNCTION:
+        ((YR_OBJECT_FUNCTION*) parent)->return_obj = obj;
+        break;
     }
   }
 
@@ -192,34 +212,39 @@ int yr_object_function_create(
 
   if (parent != NULL)
   {
+    // The parent of a function must be a structure.
+
     assert(parent->type == OBJECT_TYPE_STRUCTURE);
 
     // Try to find if the structure already has a function
-    // with that name. In that case this is a function oveload.
+    // with that name. In that case this is a function overload.
 
     f = (YR_OBJECT_FUNCTION*) yr_object_lookup_field(parent, identifier);
 
+    // Overloaded functions must have the same return type.
+
     if (f != NULL && return_type != f->return_obj->type)
       return ERROR_WRONG_RETURN_TYPE;
   }
 
   if (f == NULL) // Function doesn't exist yet
   {
-    // Let's create the result object first
-
-    FAIL_ON_ERROR(yr_object_create(return_type, "result", NULL, &return_obj));
-
-    FAIL_ON_ERROR_WITH_CLEANUP(
+    FAIL_ON_ERROR(
         yr_object_create(
             OBJECT_TYPE_FUNCTION,
             identifier,
             parent,
-            &o),
-        yr_object_destroy(return_obj));
+            &o));
+
+    FAIL_ON_ERROR_WITH_CLEANUP(
+        yr_object_create(
+            return_type,
+            "result",
+            o,
+            &return_obj),
+        yr_object_destroy(o));
 
     f = (YR_OBJECT_FUNCTION*) o;
-    f->return_obj = return_obj;
-    f->return_obj->parent = (YR_OBJECT*) f;
   }
 
   for (i = 0; i < MAX_OVERLOADED_FUNCTIONS; i++)
@@ -432,7 +457,7 @@ YR_OBJECT* _yr_object_lookup(
   {
     i = 0;
 
-    while(*p != '\0' && *p != '.' && *p != '[' && i < sizeof(str))
+    while(*p != '\0' && *p != '.' && *p != '[' && i < sizeof(str) - 1)
     {
       str[i++] = *p++;
     }
@@ -470,7 +495,7 @@ YR_OBJECT* _yr_object_lookup(
       }
       else if (*p >= '0' && *p <= '9')
       {
-        index = strtol(p, (char**) &p, 10);
+        index = (int) strtol(p, (char**) &p, 10);
       }
       else if (*p == '"')
       {
@@ -544,7 +569,6 @@ int yr_object_copy(
   YR_OBJECT* copy;
   YR_OBJECT* o;
 
-  YR_ARRAY_ITEMS* array_items;
   YR_STRUCTURE_MEMBER* structure_member;
   YR_OBJECT_FUNCTION* func;
   YR_OBJECT_FUNCTION* func_copy;
@@ -609,24 +633,27 @@ int yr_object_copy(
 
     case OBJECT_TYPE_ARRAY:
 
-      array_items = ((YR_OBJECT_ARRAY*) object)->items;
+      yr_object_copy(
+        ((YR_OBJECT_ARRAY *) object)->prototype_item,
+        &o);
 
-      for (i = 0; i < array_items->count; i++)
-      {
-        if (array_items->objects[i] != NULL)
-        {
-          FAIL_ON_ERROR_WITH_CLEANUP(
-              yr_object_copy(array_items->objects[i], &o),
-              yr_object_destroy(copy));
-
-          FAIL_ON_ERROR_WITH_CLEANUP(
-                yr_object_array_set_item(copy, o, i),
-                yr_free(o);
-                yr_object_destroy(copy));
-        }
-      }
+      ((YR_OBJECT_ARRAY *)copy)->prototype_item = o;
 
       break;
+
+    case OBJECT_TYPE_DICTIONARY:
+
+      yr_object_copy(
+        ((YR_OBJECT_DICTIONARY *) object)->prototype_item,
+        &o);
+
+      ((YR_OBJECT_DICTIONARY *)copy)->prototype_item = o;
+
+      break;
+
+    default:
+      assert(FALSE);
+
   }
 
   *object_copy = copy;
@@ -750,6 +777,8 @@ YR_OBJECT* yr_object_dict_get_item(
     int flags,
     const char* key)
 {
+  int i;
+
   YR_OBJECT* result = NULL;
   YR_OBJECT_DICTIONARY* dict;
 
@@ -759,7 +788,7 @@ YR_OBJECT* yr_object_dict_get_item(
 
   if (dict->items != NULL)
   {
-    for (int i = 0; i < dict->items->used; i++)
+    for (i = 0; i < dict->items->used; i++)
     {
       if (strcmp(dict->items->objects[i].key, key) == 0)
         result = dict->items->objects[i].obj;
@@ -1042,10 +1071,11 @@ int yr_object_set_string(
     if (string_obj->value == NULL)
       return ERROR_INSUFICIENT_MEMORY;
 
-    string_obj->value->length = len;
+    string_obj->value->length = (uint32_t) len;
     string_obj->value->flags = 0;
 
     memcpy(string_obj->value->c_string, value, len);
+    string_obj->value->c_string[len] = '\0';
   }
   else
   {
@@ -1067,7 +1097,7 @@ YR_OBJECT* yr_object_get_root(
   return o;
 }
 
-void yr_object_print_data(
+YR_API void yr_object_print_data(
     YR_OBJECT* object,
     int indent,
     int print_identifier)
@@ -1077,8 +1107,9 @@ void yr_object_print_data(
   YR_STRUCTURE_MEMBER* member;
 
   char indent_spaces[32];
+  int i;
 
-  indent = yr_min(indent, sizeof(indent_spaces));
+  indent = yr_min(indent, sizeof(indent_spaces) - 1);
 
   memset(indent_spaces, '\t', indent);
   indent_spaces[indent] = '\0';
@@ -1098,13 +1129,14 @@ void yr_object_print_data(
     case OBJECT_TYPE_STRING:
       if (((YR_OBJECT_STRING*) object)->value != NULL)
       {
+        size_t l;
         printf(" = \"");
 
-        for (int i = 0; i < ((YR_OBJECT_STRING*) object)->value->length; i++)
+        for (l = 0; l < ((YR_OBJECT_STRING*) object)->value->length; l++)
         {
-          char c = ((YR_OBJECT_STRING*) object)->value->c_string[i];
+          char c = ((YR_OBJECT_STRING*) object)->value->c_string[l];
 
-          if (isprint(c))
+          if (isprint((unsigned char) c))
             printf("%c", c);
           else
             printf("\\x%02x", (unsigned char) c);
@@ -1140,7 +1172,7 @@ void yr_object_print_data(
 
       if (array_items != NULL)
       {
-        for (int i = 0; i < array_items->count; i++)
+        for (i = 0; i < array_items->count; i++)
         {
           if (array_items->objects[i] != NULL)
           {
@@ -1158,7 +1190,7 @@ void yr_object_print_data(
 
       if (dict_items != NULL)
       {
-        for (int i = 0; i < dict_items->used; i++)
+        for (i = 0; i < dict_items->used; i++)
         {
           printf("\n%s\t%s", indent_spaces, dict_items->objects[i].key);
           yr_object_print_data(dict_items->objects[i].obj, indent + 1, 0);
